"""
The docker tool provides ``build``, ``push`` and ``publish`` commands.
"""

import logging

from menhir.tool import Tool
from menhir.tool_utils import OK, FAIL, NOTHING_TO_DO, tool_env, working_dir

log = logging.getLogger(__name__)


def tool():
    return Docker()


class Docker(Tool):

    def name(arg):
        return "docker"

    def is_using_tool(tool, path):
        from os.path import exists, join
        path = join(path, 'Dockerfile')
        return exists(path)

    def dependencies(tool, path):
        return []

    def add_arg_parser(tool, parser):
        parser.description = "Commands to build and push docker images."
        parsers = parser.add_subparsers(help="Docker commands", dest='phase')
        parsers.add_parser(
            'build',
            help='Build a docker image from a Dockerfile'
        )
        parsers.add_parser(
            'push',
            help='Push a docker image to a remote repository'
        )
        parsers.add_parser(
            'publish',
            help='Build and push a docker image to a remote repository'
        )

    def execute_build_phase(tool, path, info, args,):
        """Execute a build phase."""
        from os.path import exists, join
        from menhir.tool_utils import run_if

        phase_name = args.phase

        dockerfile = join(path, 'Dockerfile')
        repo = info['config'].get('docker', {}).get('repository')

        if not exists(dockerfile):
            log.debug(
                'No Dockerfile %(dockerfile)s',
                {'dockerfile': dockerfile})
            return NOTHING_TO_DO

        changed = info.get('changed') or info.get('changed_dependents')

        if phase_name == 'build':
            with run_if(changed, phase_name, path) as flag:
                if flag:
                    return docker_build(path, args)
                return OK

        elif phase_name == 'push':
            with run_if(changed, phase_name, path) as flag:
                if flag:
                    return docker_push(repo, path, info, args)
                return OK

        elif phase_name == 'publish':
            with run_if(changed, phase_name, path) as flag:
                if flag:
                    res = docker_build(path, args)
                    if res != OK:
                        return res
                    return docker_push(repo, path, info, args)
                return OK
        else:
            return NOTHING_TO_DO


def docker_build(path, args):
    from menhir.project import project_name_for_dir
    from menhir.tool_utils import call, package_script
    log.info('Running docker-build in %s', path)

    project_name = project_name_for_dir(path)

    env = tool_env()
    env['MENHIR_TAG'] = project_name

    with package_script("/tools/docker/docker-build.sh") as f:
        with working_dir(path):
            return call([f.name], env=env,)


def docker_push(repo, path, info, args):
    from menhir.project import branch, project_name_for_dir, image
    from menhir.tool_utils import call, package_script, slugify
    log.info('Running docker-push in %s', path)

    project_name = project_name_for_dir(path)
    current_branch = branch()
    tag = project_name
    sha_tag = image(info)
    if not sha_tag:
        log.error('No remote repository configured to push to.')
        return FAIL
    branch_tag = "%s:%s" % (
        sha_tag.split(':')[0],
        slugify(current_branch, length=40),
    )

    env = tool_env()
    env['MENHIR_TAG'] = tag
    env['MENHIR_BRANCH_TAG'] = branch_tag
    env['MENHIR_SHA_TAG'] = sha_tag

    with package_script("/tools/docker/docker-push.sh") as f:
        with working_dir(path):
            return call([f.name], env=env,)



"""Terraform tool.

Provides ``plan``, ``apply`` and ``destroy`` tasks for terrform, managing
remote-state.

Configuration is under the ``terraform`` key.

The ``targets`` key contains a dict for each target directory.  Each
infra directory is configured as a dict under the directory's name.

Arguments
---------

A directory can be configures with ``args``, a list containing the names
of values it expects to be passed on the command line.


Vars passed to Terraform
------------------------

The ``values`` key configures a list of names for values to be passed
to terraform.

``values`` can contain names from ``args`` and the following:

project        the name of the project
branch         the current branch name
branch-slug    the current branch name, sanitised
sha            the current commit sha as a hex string
sha-mod-1024   the current as a decimal modulo 1024

The ``values`` list can also contain items of the form ``name=value``,
which will pass the var ``name`` to terraform, with the value
calculated according to ``value`` from the list above.

Remote Config
-------------

The remote state is configured independently of the target direcotry
using the `remote-config` key.  The value contains a `backend` string
specifying the terraform backend, and a `backend-config` dict
specifying the backend specific configuration.

The `remote-config` key is also supported at the target directory
level.  For consistency, usage at this level is not recommended.

The only supported backend at present is ``s3``.  It expects
``bucket``, ``key`` and ``region`` values.  The `key`
configuration value is a format string, with can use ``project`` and
``component`` arguments.

In each target directory config, the ``component`` list of values is
used to construct a string to identify the remote state configuration.
``component`` defaults to ``args``.

"""
import logging

from menhir.tool import Tool
from menhir.tool_utils import FAIL, NOTHING_TO_DO, OK, tool_env, working_dir
from menhir.utils import multi, method

log = logging.getLogger(__name__)


def tool():
    return Terraform()


class Terraform(Tool):

    def is_using_tool(tool, path):
        from os.path import exists, join
        path = join(path, 'infra')
        return exists(path)

    def dependencies(tool, path):
        return []

    def add_arg_parser(tool, parser):
        parsers = parser.add_subparsers(
            help="Terraform commands",
            dest='phase'
        )
        p = parsers.add_parser(
            'apply',
            help='Apply a terraform infrastructure directory'
        )
        p.add_argument('directory', nargs=1)
        p.add_argument('args', nargs="*")
        p = parsers.add_parser(
            'plan',
            help='Plan a terraform infrastructure directory'
        )
        p.add_argument('directory', nargs=1)
        p.add_argument('args', nargs="*")
        p = parsers.add_parser(
            'destroy',
            help='Destroy a terraform infrastructure directory'
        )
        p.add_argument('directory', nargs=1)
        p.add_argument('args', nargs="*")

    def execute_build_phase(tool, path, info, args):
        """Execute a build phase."""
        from os.path import exists, join

        config = info['config'].get('terraform', {})
        log.debug('terraform config: %(config)s', {'config': config})
        infra_dir = join(path, config.get('source-dir', 'infra'))

        if not exists(infra_dir):
            log.debug('No infra directory in %(path)s', {'path': path})
            return NOTHING_TO_DO

        return task(path, info, config, args)


def task(path, info, config, args):
    """Execute a task."""
    return action(args.phase, path, info, config, args)


def action(action, path, info, config, args):
    """Invoke a terraform action on the given path."""
    from functools import reduce
    from os import getenv
    from os.path import exists, join
    from menhir.project import project_name_for_dir
    from menhir.tool_config import aliased_value_array, value_array
    from menhir.tool_utils import call
    from menhir.utils import deep_merge

    directory = args.directory[0]
    changed = info.get('changed') or info.get('changed_dependents')
    project_dir = join(path, 'infra', args.directory[0])
    project_dir_exists = exists(project_dir)

    run = project_dir_exists and (changed or action == 'destroy')
    log_run(run, args.phase, directory, path)

    if not run:
        return NOTHING_TO_DO

    dconfig = config.get('targets', {}).get(directory, {})
    dargs = dconfig.get('args', [])
    dcomponent = dconfig.get('component', dargs)
    dvalues = dconfig.get('values', [])
    remote_config = config.get('remote-config')
    dremote_config = dconfig.get('remote-config', {})
    remote_config = deep_merge(remote_config, dremote_config)

    if len(args.args or []) < len(dargs):
        log.error(
            'Not enough arguments to terraform apply %(directory)s: '
            'expected: %(expected)s, got: %(got)s',
            {
                'directory': directory,
                'expected': dargs,
                'got': args.args,
            }
        )
        return FAIL

    arg_dict = dict(zip(dargs, args.args))

    component = "-".join(value_array(dcomponent, info, path, arg_dict))

    def format_value(x):
        return '%s="%s"' % x

    values = list(map(
        format_value,
        aliased_value_array(dvalues, info, path, arg_dict),
    ))

    def extend_var(x, y):
        x.extend(['-var', y])
        return x

    var_args = reduce(extend_var, values, [],)

    project_name = project_name_for_dir(path)

    env = tool_env()

    tf_log = getenv('TF_LOG')
    if tf_log:
        env['TF_LOG'] = tf_log

    with working_dir(project_dir):
        if remote_state(remote_config, project_name, component):
            return FAIL
        res = call(["terraform", "get"], env=env,)
        if res != OK:
            return res
        log.info("terraform %s %s", action, var_args)
        return call(["terraform", action] + var_args, env=env,)


def log_run(cond, task, directory, path):
    txt = None
    if cond:
        txt = 'Run %(task)s in %(path)s'
    else:
        txt = 'Not run %(task)s in %(path)s'

    log.info(txt, {'task': task, 'directory': directory, 'path': path})


def remote_state(remote_config, project, component):
    import shutil
    if not remote_config:
        return None

    backend = remote_config.get('backend')
    if not backend:
        log.warning('remote-state configured, but no backend specified')
        return None

    backend_config = remote_config.get('backend-config')
    if not backend_config:
        log.warning('remote-state configured, but no backend_config specified')
        return None

    shutil.rmtree('.terraform', ignore_errors=True)
    return remote_state_backend(backend, backend_config, project, component)


@multi
def remote_state_backend(backend, backend_config, project, component):
    """Configure a remote backend."""
    return backend


@method(remote_state_backend, 's3')
def remote_state_s3(backend, backend_config, project, component):
    from os import getenv
    from os.path import join
    from subprocess import call

    bucket = backend_config.get('bucket')
    if not bucket:
        log.warning('terraform remote-state s3: no bucket specified')
        return None

    key = backend_config.get('key', "%(project)-%(component)")

    component = component or 'config'
    key = key % {'project': project, 'component': component}
    key = join(key, "terraform.tfstate")
    region = backend_config.get('region', "us-east-1")

    log.info(
        'State from s3://%(bucket)s/%(key)s in %(region)s',
        {
            'bucket': bucket,
            'key': key,
            'region': region,
        }
    )

    env = tool_env()

    tf_log = getenv('TF_LOG')
    if tf_log:
        env['TF_LOG'] = tf_log

    cmd = [
            "terraform", "remote", "config",
            "-backend=s3",
            "-backend-config=bucket=%s" % bucket,
            "-backend-config=key=%s" % key,
            "-backend-config=region=%s" % region,
        ]

    return call(cmd, env=env)



def find_file_in_parents(file_name, base_dir=None):
    """From base_dir, look for a file in parent directories.

    Returns the full oath to the file found.

    `base_dir` defaults to the current directory.
    """
    import os
    import os.path

    dir = base_dir or os.getcwd()
    while True:
        f = os.path.join(dir, file_name)
        if os.path.exists(f):
            return f
        else:
            parent_dir = os.path.dirname(dir)
            if dir == parent_dir:  # if dir is root dir
                return None
            else:
                dir = parent_dir


def load_yaml(path):
    """Load the yaml at the specified path.

    If the path doesn't exist, return None.
    """
    from os.path import exists
    import yaml
    if exists(path):
        with open(path, 'r') as stream:
            return yaml.load(stream)


def load_json(json_file):
    import json
    import sys
    try:
        if json_file != "-":
            with open(json_file) as f:
                return json.load(f)
        return json.load(sys.stdin)
    except Exception:
        print('Failed to load JSON file.')
        raise


def load_encrypted(path, key):
    """Load an encrypted file into memory, decrypting it."""
    import subprocess
    cmd = ["openssl", "aes-256-cbc", "-a", "-k", key,  "-in", path, "-d"]
    ps = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    return ps.communicate()[0].decode('utf-8')


def directory_graph(base_dir):
    """Return a graph of directories."""
    from os import walk
    from os.path import join, normpath

    stop_dirs = {'__pycache__'}
    graph = {}
    for root, subdirs, files in walk(base_dir):
        prune = [subdir for subdir in subdirs
                 if subdir in stop_dirs or subdir.startswith('.')]
        for subdir in prune:
            subdirs.remove(subdir)
        root = normpath(root)
        for subdir in subdirs:
            graph[root] = [normpath(join(root, subdir)) for subdir in subdirs]
    return graph


def paths_between(parent, child):
    """Return a list of path segments between parent and child."""
    from os import pathsep
    from os.path import abspath, join, normpath
    from menhir.utils import reductions
    if parent == child:
        return []
    parent = normpath(abspath(parent))
    child = normpath(abspath(child))
    assert child.startswith(parent)
    paths = child[len(parent)+1:].split(pathsep)
    paths.pop()
    return [parent] + list(reductions(join, paths, parent))



# menhir configuration
import logging

log = logging.getLogger(__name__)
ROOT_FILE = 'menhir_root.yaml'


def find_root():
    """Based on the current directory, look for a .menhir_root.yaml file."""
    from .fileutils import find_file_in_parents
    root = find_file_in_parents(ROOT_FILE)
    if root is None:
        raise FileNotFoundError(
            'No root file found in parent directories',
            ROOT_FILE
        )
    return root


def root_config(root_file=ROOT_FILE):
    """Return the root menhir configuration."""
    from os.path import exists
    import yaml

    if exists(root_file):
        with open(root_file, 'r') as stream:
            config = yaml.load(stream)
    else:
        config = default_root_config()
    return parse_root_config(config)


def parse_root_config(config):
    tools = config.get('tools')
    config['tools'] = {tool: load_tool(tool) for tool in tools}
    return config


def default_root_config():
    return {
        'tools': default_tools(),
    }


def default_tools():
    from menhir.tools import all_tools
    return all_tools


def describe_repository(config):
    """Parse the root menhir configuration."""
    import os
    from os.path import relpath
    from menhir import gitutils
    from menhir.fileutils import directory_graph
    from menhir.graph import root_nodes

    git_root = gitutils.find_root()
    rel_git_root = relpath(git_root)

    tool_impls = config['tools'].values()

    # build the dependency graph
    fs_graph = directory_graph(rel_git_root)
    infos = project_infos(fs_graph, rel_git_root, tool_impls)
    graph = project_graph(infos)
    roots = root_nodes(graph, infos)

    load_project_config(fs_graph, rel_git_root, infos)

    def is_child(x, y):
        from os.path import abspath, normpath
        return normpath(abspath(y)).startswith(normpath(abspath(x)))

    roots = set(filter(lambda x: is_child('.', x), roots))

    cwd = os.getcwd()
    git_prefix = cwd[len(git_root)+1:]

    config['tool_impls'] = tool_impls
    config['project_infos'] = infos
    config['dependencies'] = graph
    config['root_projects'] = roots
    config['git_root'] = git_root
    config['git_prefix'] = git_prefix
    return config


def defaults_config(dir_path):
    """Return the defaults configuration in dir_path."""
    from os.path import join
    from menhir.fileutils import load_yaml
    return load_yaml(join(dir_path, 'menhir-defaults.yaml')) or {}


def project_config(dir_path):
    """Return the project configuration in dir_path."""
    from os.path import join
    from menhir.fileutils import load_yaml
    return load_yaml(join(dir_path, 'menhir-config.yaml')) or {}


def load_tool(tool_name):
    """Load the tool implementation for the given tool name."""
    tool = \
        require_tool(tool_name) or \
        require_tool('menhir.tools.%s' % tool_name)
    if not tool:
        raise Exception('configuration_error', 'Tool not found %s', tool_name)
    return tool


def require_tool(tool_path):
    from importlib import import_module
    try:
        m = import_module(tool_path)
        return m.tool()
    except Exception as e:
        log.debug('Failed to import: %s, "%s"', tool_path, e)


def project_infos(graph, root, tools):
    """Return a map of project root path to project info dict."""
    from menhir.graph import dfs_visit

    infos = {}

    def pre_f(state, path):
        used = list(filter(lambda t: t.is_using_tool(path), tools))
        if len(used):
            infos[path] = infos.get(path, {})
            infos[path]['tools'] = used
        return state

    def post_f(state, path, child_states):
        return state

    state = {}
    dfs_visit(graph, root, state, pre_f, post_f,)

    return infos


def load_project_config(graph, root, infos):
    """Load project configuration."""
    from menhir.graph import dfs_visit
    from menhir.utils import deep_merge

    def pre_f(defaults, path):
        from copy import deepcopy
        defaults = deep_merge(defaults, defaults_config(path))
        config = deepcopy(defaults)
        config = deep_merge(config, project_config(path))
        if path in infos:
            infos[path]['config'] = config
        return defaults

    def post_f(defaults, path, child_defaults):
        return defaults

    defaults = {}
    dfs_visit(graph, root, defaults, pre_f, post_f,)

    return infos


def infos_config(base_dir, tools):
    """Return a map of project root path to project info dict."""
    from menhir.fileutils import directory_graph
    from menhir.graph import dfs_visit

    infos = {}

    def config_with_defaults(defaults, path):
        from copy import deepcopy
        defaults.update(defaults_config(path))
        config = deepcopy(defaults)
        config.update(project_config(path))
        if path in infos:
            infos[path]['config'] = config
        return defaults

    def tooling(path):
        used = list(filter(lambda t: t.is_using_tool(path), tools))
        if len(used):
            infos[path] = infos.get(path, {})
            infos[path]['tools'] = used

    def pre_f(state, path):
        tooling(path)
        state = config_with_defaults(state, path)
        return state

    def post_f(state, path, child_states):
        return state

    graph = directory_graph(base_dir)
    state = {}
    dfs_visit(graph, base_dir, state, pre_f, post_f,)

    return infos


def project_graph(roots):
    """Return a project dependency graph."""
    from functools import reduce

    return {
        root: reduce(
            lambda x, y: x.union(y or set()),
            [
                set(tool.dependencies(root))
                for tool in info['tools']
            ],
            set()
        )
        for root, info in roots.items()
        }



"""Menhir command line script.

The script is the main user interface for menhir.
"""
from __future__ import print_function  # NOQA

import logging

from menhir import __version__
from menhir.config import default_root_config, parse_root_config
from menhir.utils import multi, method


log = logging.getLogger(__name__)


def main():
    from menhir import root_config
    config = root_config()
    args = parser(config).parse_args()

    if args.verbose and args.log_level != 'DEBUG':
        args.log_level = 'INFO'
    logging.basicConfig(level=args.log_level)

    write_version()
    if args.version:
        print('Menhir %s' % __version__)
    elif args.menhir_command:
        command(config, args,)


def parser(config=parse_root_config(default_root_config())):
    import argparse
    parser = argparse.ArgumentParser(description="""
Menhir is an extensible build tool, that recognises dependencies
between sub-projects in a repository.
""")
    parser.add_argument(
        "--version",
        default=False,
        action='store_const',
        const=True,
        help="Print menhir version and exit",
    )
    parser.add_argument("--commit", help="""
Specify commit range to work on.
This can be specifed as a single commit or a commit range, e.g.
3f8e5bb, 3f8e5bb..cfa3762, 3f8e5bb^..head.
""")
    parser.add_argument(
        "-v", "--verbose",
        default=False,
        action='store_const', const=True,
        help='Provide verbose output (shortcut for log-levl INFO)',
    )
    parser.add_argument("--circleci", default=False,
                        action='store_const', const=True)
    parser.add_argument(
        "--log-level", default='WARNING',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        help="Set the log level to output"
    )

    subparsers = parser.add_subparsers(
        help='sub-commands',
        dest='menhir_command'
    )

    parser_exec = subparsers.add_parser(
        'info',
        help="Print project information")

    parser_exec = subparsers.add_parser('exec', help='execute a build tool')
    parser_exec.add_argument("--only-if-changed", default=False,
                             action='store_const', const=True)

    tool_parsers = parser_exec.add_subparsers(help='tools', dest='tool')
    for tool_name, tool in config['tools'].items():
        tool_parser = tool_parsers.add_parser(
            tool_name,
        )
        tool.add_arg_parser(tool_parser)

    return parser


@multi
def command(root_config, args):
    return args.menhir_command


@method(command, 'info')
def info(root_config, args):
    from pprint import pprint
    from menhir import describe_repository
    config = describe_repository(root_config)
    pprint(config)


@method(command, 'exec')
def exec_tool(root_config, args):
    import sys
    from menhir import (
        apply_tool,
        describe_repository,
        update_for_changed_files,
    )

    config = describe_repository(root_config)
    tool = root_config['tools'][args.tool]

    project_infos = config['project_infos']
    dependencies = config['dependencies']
    root_projects = config['root_projects']
    git_prefix = config['git_prefix']

    files = changed_files(config, args.circleci, args.commit)
    update_for_changed_files(
        dependencies, project_infos, root_projects, git_prefix, files
    )

    res = apply_tool(
        dependencies,
        project_infos,
        root_projects,
        tool,
        args,
    )

    if res.get('failed'):
        print('Build tool "%s" failed' % args.tool, file=sys.stderr)
        sys.exit(1)


def changed_files(config, circleci, commit):
    from menhir.gitutils import changed_files, repo, uncommited_files
    git_repo = repo(config['git_root'])
    uncommited = uncommited_files(git_repo)['all']
    if uncommited and not commit:
        files = uncommited
        log.info("Uncommited files")
    else:
        from_commit, to_commit = commit_range(
            commit, circleci, config['git_root']
        )
        log.info("Commit range: %s..%s", from_commit, to_commit)

        files = changed_files(from_commit, to_commit)['all']

    return files


def commit_range(commit_arg, circleci, root):
    import os
    from menhir.gitutils import head_commit, repo
    if commit_arg:
        return commits_from_range(commit_arg)
    if circleci:
        compare_url = os.getenv('CIRCLE_COMPARE_URL')
        log.debug('Deriving commit range from %s', compare_url)
        if compare_url:
            commit_range = compare_url.rsplit('/', 1)[1]
            return commits_from_range(commit_range)
        sha = os.getenv('CIRCLE_SHA1')
        return sha+"^", sha

    r = repo(root)
    commit = head_commit(r)
    return str(commit)+"^", str(commit)


def commits_from_range(commit_range):
    log.debug('commits_from_range: %s', commit_range)
    commits = commit_range.split('..')
    if len(commits) == 2:
        from_commit = commits[0].replace('.', '')
        to_commit = commits[1].replace('.', '')
        log.debug('commits_from_range: from %s to %s', from_commit, to_commit)
        return from_commit, to_commit
    log.debug('commits_from_range: single commit %s', commit_range)
    return commit_range+"^", commit_range


def write_version():
    from menhir import __version__
    with open('.menhir-version', 'w+') as f:
        f.write(__version__)



try:
    from functools import lru_cache
except:
    from functools32 import lru_cache


def project_name():
    """Return the name of the project in the current directory."""
    from os import getcwd
    return project_name_for_dir(getcwd())


@lru_cache()
def project_name_for_dir(path):
    """Return the name of the project in the current directory."""
    from os.path import abspath, basename, exists, join
    fname = join(path, '.project-name')
    if exists(fname):
        with open(fname, 'r') as f:
            return f.read().strip()
    return basename(abspath(path))


@lru_cache()
def sha():
    from menhir.gitutils import head_commit, sha
    return sha(head_commit(repo()))


@lru_cache()
def branch():
    from menhir.gitutils import branch
    return branch(repo())


@lru_cache()
def repo():
    from menhir.gitutils import repo
    return repo()


def image(info):
    repo = info.get('config', {}).get('docker', {}).get('repository')
    if repo:
        return '%s/%s:%s' % (
            repo,
            project_name(),
            sha()
        )



# GIT utilities
import logging
from contextlib import contextmanager

from git import Repo

log = logging.getLogger(__name__)


def find_root():
    """Based on the current directory, look for a .menhir_root.yaml file."""
    from os.path import dirname
    from .fileutils import find_file_in_parents
    root = find_file_in_parents(".git")
    if root is None:
        raise FileNotFoundError(
            'No git repository in parent directories',
            ".git"
        )
    return dirname(root)


def repo(base_dir=None):
    """Return a repository object."""
    base_dir = base_dir or find_root()
    return Repo(base_dir)


def commit(repo, commitish):
    """Return a commit object."""
    return repo.commit(commitish)


def commit_cached(repo, *args):
    """Commit cached files"""
    # return repo.index.commit(message, **kwargs)
    repo.git.commit(*args)


def head_commit(repo):
    """Return a commit object."""
    return repo.head.commit


def diff(start, end):
    """Diff for the commit range."""
    return diff_paths(start.diff(end))


def dirty_files(repo):
    return diff_paths(repo.index.diff(None))


def staged_files(repo):
    return diff_paths(repo.index.diff("HEAD"))


def uncommited_files(repo):
    dirty = repo.index.diff(None)
    dirty.extend(repo.index.diff("HEAD"))
    return diff_paths(dirty)


def unstaged_files(repo):
    dirty = repo.index.diff(None)
    dirty.extend(repo.index.diff("HEAD"))
    return diff_paths(dirty)


def diff_paths(diff):
    rpaths = [p.a_path for p in diff.iter_change_type('R')]
    rpaths.extend([p.b_path for p in diff.iter_change_type('R')])

    diffs = {
        'added': [p.b_path for p in diff.iter_change_type('A')],
        'deleted': [p.a_path for p in diff.iter_change_type('D')],
        'renamed': rpaths,
        'modified': [p.b_path for p in diff.iter_change_type('M')],
    }
    all_paths = []
    for i in ['added', 'deleted', 'renamed', 'modified']:
        all_paths.extend(diffs[i])
    diffs['all'] = all_paths
    return diffs


def changed_files(start, end):
    from menhir import gitutils
    repo = gitutils.repo()
    start_commit = gitutils.commit(repo, start)
    end_commit = gitutils.commit(repo, end)
    changed_files = gitutils.diff(start_commit, end_commit)
    return changed_files


def sha(commit):
    """Return the sha hex for a commit."""
    return commit.hexsha


def branch(repo):
    """Return the branch name."""
    return repo.active_branch.name


@contextmanager
def staged_as_committed(repo):
    staged_changes = staged_files(repo)
    log.debug('staged_changes %s', staged_changes)

    commit_changes = False
    if len(staged_changes['all']):
        commit_cached(repo, "-m", "temp", "-n")
        commit_changes = True

    try:
        v = yield
    finally:
        if commit_changes:
            repo.git.reset("HEAD^", soft=True)
    return v


@contextmanager
def stashed_changes(repo):
    uncommitted_changes = uncommited_files(repo)
    log.debug('uncommitted_changes %s', uncommitted_changes)
    stash_changes = False

    if len(uncommitted_changes['all']):
        stash_changes = True
        repo.git.stash('save', "--quiet", "--include-untracked")
    try:
        v = yield
    finally:
        if stash_changes:
            repo.git.stash('pop', "--quiet")
    return v



"""
Menhir is a project automation tool that is designed to work with
a monrepo. It is extensible using python packages.
"""
import logging

from menhir.config import describe_repository, root_config
from menhir.tool_utils import OK, NOTHING_TO_DO

__version__ = "0.0.22"

log = logging.getLogger(__name__)


def apply_tool(dependencies, dirs, roots, tool, args):
    """Apply phase_name to each project.

    Projects are visited in dependency order.  A full traversal is
    made.  It is up to each tool to decide whether to run, based on
    the changed state in the project info.
    """
    from functools import reduce
    from menhir.graph import dfs_visit

    def pre_f(state, path):
        return state

    def merge_state(x, y):
        log.debug('merge_state: %s %s', x, y)
        x['failed'] = x.get('failed') or y.get('failed')
        x['seen'] = x['seen'].union(y['seen'])
        return x

    def post_f(state, path, child_states):
        log.debug('apply_tool post_f %s %s', path, state)
        info = dirs[path]
        if not state.get('failed') and path not in state['seen']:
            res = apply_tool_to_path(path, info, tool, args)
            if res['status'] == 'fail':
                log.warning('%s failed in %s', tool, path)
                state['failed'] = True

        seen = state['seen']
        seen.add(path)
        state['seen'] = seen
        return state

    state = {'seen': set()}
    child_states = [
        dfs_visit(
            dependencies,
            root,
            state,
            pre_f,
            post_f,
            child_state_f=lambda x: x  # shared state
        )
        for root in roots
    ]

    log.debug('child_states: %s', child_states)
    return reduce(merge_state, child_states, state)


def apply_tool_to_path(path, info, tool, args):
    if tool not in set(info['tools']):
        log.debug('Tool %s not used in %s', tool, path)
        return NOTHING_TO_DO
    res = tool.execute_build_phase(path, info, args)
    log.debug('Execute tool %s in %s returns %s', tool.name(), path, res)
    if res['status'] == 'fail':
        log.debug('Tool %s in %s failed %s', tool.name(), path, res)
        res['tool'] = tool.name()
        res['args'] = args
        res['path'] = path
        return res
    return OK


def update_for_changed_files(
        dependencies, project_infos, root_projects, prefix, files
):
    """Update project infos with changed status, given changed files.

    Updates project_infos with `changed`, `changed_dependees` and
    `changed_dependents` fields, based on the list of changed files.

    `prefix` allows the specification of a path prefix, to add to each
    of the project dirs, before determining if they contain any of the
    files.

    Returns a list of changed projects.

    """
    changed_projects = projects_with_files(project_infos, files, prefix)
    mark_changed(project_infos, changed_projects)
    propagate_changed(dependencies, project_infos, root_projects)
    return changed_projects


def projects_with_files(roots, files, root_prefix):
    """Filter roots that contain any of files paths.

    The roots are each prefixed with root_prefix, before comparison.
    """
    from os.path import join, normpath
    log.debug('projects_with_files:')
    log.debug(' roots: %s', roots)
    log.debug(' files: %s', files)
    roots = set(roots)
    changed_roots = set()

    for path in files:
        if len(roots):
            for root in roots.copy():
                prefix = normpath(join(root_prefix, root)) if root_prefix \
                         else root
                log.debug('prefix: %s, path: %s', prefix, path)
                if path.startswith(prefix) or (
                        prefix == '.' and not path.startswith('..')
                ):
                    changed_roots.add(root)
                    roots.remove(root)
        else:
            break
    return changed_roots


CHANGED = 'changed'
CHANGED_DEPENDENTS = 'changed_dependents'
CHANGED_DEPENDEES = 'changed_dependees'


def mark_changed(projects, changed_projects):
    for project in changed_projects:
        projects[project][CHANGED] = True


def propagate_changed(dependencies, infos, root_projects):
    """Propagate changed info to dependees and dependents.

    `changed` nodes propagate `changed_dependees` down the graph.
    `changed` nodes propagate `changed_dependents` up the graph.
    """
    from menhir.graph import dfs_visit

    def pre_f(state, node):
        info = infos[node]
        if state.get(CHANGED_DEPENDEES):
            info[CHANGED_DEPENDEES] = True
        if info.get(CHANGED):
            state[CHANGED_DEPENDEES] = True
        log.debug('info pre: %s', info)
        return state

    def post_f(state, node, child_states):
        from copy import copy
        info = infos[node]
        for child_state in child_states:
            if (
                    child_state.get(CHANGED_DEPENDENTS) or
                    child_state.get(CHANGED)
            ):
                info[CHANGED_DEPENDENTS] = True
        log.debug('info post: %s', info)
        return copy(info)

    for root in root_projects:
        dfs_visit(dependencies, root, {}, pre_f, post_f)


__all__ = [
    'describe_repository',
    'root_config',
    'apply_tool',
    'projects_with_files',
    'mark_changed',
    'propagate_changed'
]
